#if !defined(TSVD_HPP)
#define TSVD_HPP

#include <Eigen/Core>
#include <Eigen/SparseCore>
#include <Eigen/Eigenvalues> 
using Eigen::Dynamic;
//typedef Eigen::Map<Eigen::MatrixXd> MapMat;
//typedef Eigen::Map<Eigen::VectorXd> MapVec;
//typedef Eigen::SparseMatrix<double> SpMatrix;

#include <SymEigsSolver.h>
#include <MatOp/DenseSymMatProd.h>
#include <MatOp/SparseSymMatProd.h>

#include <GenEigsSolver.h>
#include <MatOp/DenseGenMatProd.h>
#include <MatOp/SparseGenMatProd.h>

#include <MatOp/SVDOp.h>
using namespace Spectra;

#include <algorithm>
using std::min;
using std::max;

// Traits to obtain operation type from matrix type
template <typename Scalar, typename MatType>
struct OpTypeTraitSym
{
    typedef DenseSymMatProd<Scalar> OpType;
};

template <typename Scalar>
struct OpTypeTraitSym<Scalar, Eigen::SparseMatrix<Scalar> >
{
    typedef SparseSymMatProd<Scalar> OpType;
};

// Traits to obtain operation type from matrix type
template <typename Scalar, typename MatType>
struct OpTypeTraitGen
{
    typedef DenseGenMatProd<Scalar> OpType;
};

template <typename Scalar>
struct OpTypeTraitGen<Scalar, Eigen::SparseMatrix<Scalar> >
{
    typedef SparseGenMatProd<Scalar> OpType;
};

// Traits to obtain SVD operation type from matrix operation type
template <typename Scalar, typename MatOpType, bool selector>
struct OpTypeTraitSvd
{
    typedef SVDTallOp<Scalar, MatOpType> OpType;
};

template <typename Scalar, typename MatOpType>
struct OpTypeTraitSvd<Scalar, MatOpType, false>
{
    typedef SVDWideOp<Scalar, MatOpType> OpType;
};

namespace oia {
    namespace maths {

        // A = U S V^*
        // So A^{-1} = V S^{-1} U^*
        /*
        Index l_rank = rank();
        tmp.noalias() =  m_matrixU.leftCols(l_rank).adjoint() * rhs;
        tmp = m_singularValues.head(l_rank).asDiagonal().inverse() * tmp;
        dst = m_matrixV.leftCols(l_rank) * tmp;

        psudoinverse(A) = v * d.asDiagonal().inverse() * u.adjoint();
        trucated(A) = u * d.asDiagonal() * v.adjoint();
        */

        template <typename Scalar>
        class Tsvd
        {
        public:
            enum { Default = -1 };
            struct Options
            {
                double tolerance;
                int maximum_number_of_iterations;
            };

            Tsvd();
            void set_options(const Options& options);

            template <typename MatType>
            int tsvd_sym(MatType A, int number_of_singular_values, int number_of_left_singular_vectors, int number_of_right_singular_vectors, int Lanzcos_basis_size,
                Eigen::Matrix<Scalar, Dynamic, 1>& d, Eigen::Matrix<Scalar, Dynamic, Dynamic>& u, Eigen::Matrix<Scalar, Dynamic, Dynamic>& v,
                int& number_of_iterations, int& number_of_operations);

            template <typename MatType>
            int tsvd_gen(MatType A, int number_of_singular_values, int number_of_left_singular_vectors, int number_of_right_singular_vectors, int Lanzcos_basis_size,
                Eigen::Matrix<Scalar, Dynamic, 1>& d, Eigen::Matrix<Scalar, Dynamic, Dynamic>& u, Eigen::Matrix<Scalar, Dynamic, Dynamic>& v,
                int& number_of_iterations, int& number_of_operations);
        private:
            Options options;
        };

        template <typename Scalar>
        Tsvd<Scalar>::Tsvd()
        {
            options.tolerance = 1.0e-10;
            options.maximum_number_of_iterations = 1000;
        }

        template <typename Scalar>
        void Tsvd<Scalar>::set_options(const Options& options)
        {
            this->options = options;
        }

        /*
        A is the matrix whose truncated SVD is to be computed.
        number_of_singular_values is a number of singular values requested
        number_of_left_singular_vectors is a number of left singular vectors to be computed.
        It must satisfy: 0 <= number_of_left_singular_vectors <= number_of_singular_values
        number_of_right_singular_vectors is a number of right singular vectors to be computed.
        It must satisfy:  0 <= number_of_right_singular_vectors <= number_of_singular_values
        Lanzcos_basis_size is a number of Lanzcos basis vectors to use.  More vectors will result in faster convergence, but
        with greater memory use. Lanzcos_basis_size must satisfy: number_of_singular_values < Lanzcos_basis_size <= p where p = min(m, n).
        Default is min(p, max(2 * number_of_singular_values + 1, 20)).
        options.tolerance is the precision parameter. Default is 1e-10.
        options.maximum_number_of_iterations is a maximum number of iterations. Default is 1000.


        d is a vector of the computed singular values.
        u is an m by nu matrix whose columns contain the left singular vectors. If number_of_left_singular_vectors == 0, NULL will be returned.
        v is an n by nv matrix whose columns contain the right singular vectors. If number_of_right_singular_vectors == 0, NULL will be returned.

        number_of_converged_singular_values is a number of converged singular values.
        number_of_iterations is a number of iterations used.
        nops is a number of matrix-vector multiplications used
        */

        template <typename Scalar>
        template <typename MatType>
        int Tsvd<Scalar>::tsvd_sym(MatType A, int number_of_singular_values, int number_of_left_singular_vectors, 
            int number_of_right_singular_vectors, int Lanzcos_basis_size,
            Eigen::Matrix<Scalar, Dynamic, 1>& d, Eigen::Matrix<Scalar, Dynamic, Dynamic>& u, Eigen::Matrix<Scalar, Dynamic, Dynamic>& v,
            int& number_of_iterations, int& number_of_operations)
        {
            int number_of_columns = A.cols();
            if (Lanzcos_basis_size == Default)
            {
                Lanzcos_basis_size = min(number_of_columns, max(2 * number_of_singular_values + 1, 20));
            }

            if (number_of_left_singular_vectors == Default)
            {
                number_of_left_singular_vectors = number_of_singular_values;
            }
            if (number_of_right_singular_vectors == Default)
            {
                number_of_right_singular_vectors = number_of_singular_values;
            }

            Eigen::Matrix<Scalar, Dynamic, 1> evals;
            Eigen::Matrix<Scalar, Dynamic, Dynamic> evecs;
            int number_of_converged_singular_values;
            if (number_of_singular_values < number_of_columns)
            {
                typename OpTypeTraitSym<Scalar, MatType>::OpType op(A);
                SymEigsSolver<Scalar, LARGEST_MAGN, typename OpTypeTraitSym<Scalar, MatType>::OpType>
                    eigs(&op, number_of_singular_values, Lanzcos_basis_size);

                eigs.init();
                number_of_converged_singular_values = eigs.compute(options.maximum_number_of_iterations, options.tolerance);
                number_of_iterations = eigs.num_iterations();
                number_of_operations = eigs.num_operations();

                if (number_of_converged_singular_values < number_of_singular_values)
                {
                    // fprintf(stderr, "only %d singular values converged, less than knumber_of_singular_values = %d", 
                    //    number_of_converged_singular_values, number_of_singular_values);
                }

                number_of_left_singular_vectors = std::min(number_of_left_singular_vectors, number_of_converged_singular_values);
                number_of_right_singular_vectors = std::min(number_of_right_singular_vectors, number_of_converged_singular_values);

                evals = eigs.eigenvalues();
                evecs = eigs.eigenvectors();
            }
            else if (number_of_singular_values == number_of_columns)
            {
                int option = (number_of_left_singular_vectors == 0 && number_of_right_singular_vectors == 0) == false ?
                    Eigen::ComputeEigenvectors : Eigen::EigenvaluesOnly;
                SelfAdjointEigenSolver<MatrixXd> eigs(A, option);
                evals = eigs.eigenvalues();
                evecs = eigs.eigenvectors();
                number_of_converged_singular_values = evals.size();
                number_of_iterations = 0;
                number_of_operations = 0;

                // Sort eigenvalues and corresponding vectors.
                for (Index i = 0; i < number_of_columns - 1; ++i)
                {
                    Index k;
                    evals.segment(i, number_of_columns - i).maxCoeff(&k);
                    if (k > 0)
                    {
                        std::swap(evals[i], evals[k + i]);
                        if (option == Eigen::ComputeEigenvectors)
                            evecs.col(i).swap(evecs.col(k + i));
                    }
                }
            }

            d = evals.head(number_of_converged_singular_values);

            if (number_of_left_singular_vectors > 0)
            {
                u = evecs.leftCols(number_of_left_singular_vectors);
            }

            if (number_of_right_singular_vectors > 0)
            {
                v = evecs.leftCols(number_of_right_singular_vectors);

                // We need to make sure that singular values are nonnegative,
                // so move the sign to v.
                for (int i = 0; i < number_of_converged_singular_values; ++i)
                {
                    if (d[i] < 0)
                    {
                        d[i] = -d[i];
                        if (i < number_of_right_singular_vectors)
                        {
                            v.col(i) = -v.col(i);
                        }
                    }
                }
            }

            return 0;
        }

        template <typename Scalar>
        template <typename MatType>
        int Tsvd<Scalar>::tsvd_gen(MatType A, int number_of_singular_values, int number_of_left_singular_vectors, 
            int number_of_right_singular_vectors, int Lanzcos_basis_size,
            Eigen::Matrix<Scalar, Dynamic, 1>& d, Eigen::Matrix<Scalar, Dynamic, Dynamic>& u, Eigen::Matrix<Scalar, Dynamic, Dynamic>& v,
            int& number_of_iterations, int& number_of_operations)
        {
            int number_of_rows = A.rows();
            int number_of_columns = A.cols();

            if (Lanzcos_basis_size == Default)
            {

                int p = min(number_of_rows, number_of_columns);
                Lanzcos_basis_size = min(p, max(2 * number_of_singular_values + 1, 20));
            }

            if (number_of_left_singular_vectors == Default)
            {
                number_of_left_singular_vectors = number_of_singular_values;
            }
            if (number_of_right_singular_vectors == Default)
            {
                number_of_right_singular_vectors = number_of_singular_values;
            }

            // Operation for original matrix
            typename OpTypeTraitGen<Scalar, MatType>::OpType op_orig(A);
            int number_of_converged_singular_values;
            Eigen::MatrixXd evecs;
            if (number_of_rows > number_of_columns)
            {
                typename OpTypeTraitSvd<Scalar, OpTypeTraitGen<Scalar, MatType>::OpType, true >::OpType op(&op_orig);
                SymEigsSolver<Scalar, LARGEST_MAGN, OpTypeTraitSvd<Scalar, OpTypeTraitGen<Scalar, MatType>::OpType, true >::OpType>
                    eigs(&op, number_of_singular_values, Lanzcos_basis_size);

                eigs.init();
                int number_of_converged_singular_values = eigs.compute(options.maximum_number_of_iterations, options.tolerance);
                if (number_of_converged_singular_values < number_of_singular_values)
                {
                    // fprintf(stderr, "only %d singular values converged, less than number_of_singular_values = %d", 
                    //    number_of_converged_singular_values, number_of_singular_values);
                }

                number_of_left_singular_vectors = std::min(number_of_left_singular_vectors, number_of_converged_singular_values);
                number_of_right_singular_vectors = std::min(number_of_right_singular_vectors, number_of_converged_singular_values);

                // Copy evals to d and take the square root
                d = eigs.eigenvalues().head(number_of_converged_singular_values).array().sqrt();

                evecs = eigs.eigenvectors(std::max(number_of_left_singular_vectors, number_of_right_singular_vectors));
                number_of_iterations = eigs.num_iterations();
                number_of_operations = eigs.num_operations() * 2;
            }
            else
            {
                typename OpTypeTraitSvd<Scalar, OpTypeTraitGen<Scalar, MatType>::OpType, false >::OpType op(&op_orig);
                SymEigsSolver<Scalar, LARGEST_MAGN, OpTypeTraitSvd<Scalar, OpTypeTraitGen<Scalar, MatType>::OpType, false >::OpType>
                    eigs(&op, number_of_singular_values, Lanzcos_basis_size);

                eigs.init();
                int number_of_converged_singular_values = eigs.compute(options.maximum_number_of_iterations, options.tolerance);
                if (number_of_converged_singular_values < number_of_singular_values)
                {
                    // fprintf(stderr, "only %d singular values converged, less than number_of_singular_values = %d", 
                    //    number_of_converged_singular_values, number_of_singular_values);
                }

                number_of_left_singular_vectors = std::min(number_of_left_singular_vectors, number_of_converged_singular_values);
                number_of_right_singular_vectors = std::min(number_of_right_singular_vectors, number_of_converged_singular_values);

                // Copy evals to d and take the square root
                d = eigs.eigenvalues().head(number_of_converged_singular_values).array().sqrt();

                evecs = eigs.eigenvectors(std::max(number_of_left_singular_vectors, number_of_right_singular_vectors));
                number_of_iterations = eigs.num_iterations();
                number_of_operations = eigs.num_operations() * 2;
            }

            // Copy evecs to u or v according to the shape of A
            // If A is tall, copy evecs to v, otherwise copy to u
            if (number_of_rows > number_of_columns && number_of_right_singular_vectors > 0)
            {
                v = evecs.leftCols(number_of_right_singular_vectors);
            }
            else if (number_of_left_singular_vectors > 0)
            {
                u = evecs.leftCols(number_of_left_singular_vectors);
            }

            // Calculate the other one
            if (number_of_rows > number_of_columns)
            {
                u.resize(number_of_rows, number_of_left_singular_vectors);
                // A = UDV', A'A = VD^2V', AV = UD, ui = A * vi / di
                // evecs has already been copied to v, so we can overwrite evecs
                for (int i = 0; i < number_of_left_singular_vectors; ++i)
                {
                    evecs.col(i) /= d[i];
                    op_orig.perform_op(&evecs(0, i), &u(0, i));
                    ++number_of_operations;
                }
            }
            else {
                v.resize(number_of_columns, number_of_right_singular_vectors);
                // A = UDV', AA' = UD^2U', A'U = VD, vi = A' * ui / di
                // evecs has already been copied to u, so we can overwrite evecs
                for (int i = 0; i < number_of_right_singular_vectors; ++i)
                {
                    evecs.col(i) /= d[i];
                    op_orig.perform_tprod(&evecs(0, i), &v(0, i));
                    ++number_of_operations;
                }
            }

            return 0;
        }



    }

}

#endif // TSVD_HPP

